﻿using Microsoft.Win32;
using System;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.IO;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Threading;

namespace TestProject
{
    /// <summary>
    /// 메인 윈도우
    /// </summary>
    public partial class MainWindow : Window
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 전체 화면 여부
        /// </summary>
        private bool isFullScreen = false;

        /// <summary>
        /// 정지 여부
        /// </summary>
        private bool isPause = true;

        /// <summary>
        /// 영상 종료 여부
        /// </summary>
        private bool isClosed = false;

        /// <summary>
        /// 타임라인 이동 여부
        /// </summary>
        private bool isTimelineMoving = false;

        /// <summary>
        /// 현재 재생 인덱스
        /// </summary>
        private int currentPlayIndex = 0;

        /// <summary>
        /// 타이머
        /// </summary>
        private DispatcherTimer timer = new DispatcherTimer();

        /// <summary>
        /// 미디어 재생 리스트 컬렉션
        /// </summary>
        private ObservableCollection<MediaPlayListItem> mediaPlayCollection = new ObservableCollection<MediaPlayListItem>();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainWindow()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();
            
            this.videoBorder.AllowDrop = true;

            this.timer.Interval = TimeSpan.FromSeconds(1);

            MouseLeftButtonDown                            += (s, e) => DragMove();
            KeyDown                                        += Window_KeyDown;

            this.setFullScreenButton.Click                 += (s, e) => SetFullScreen();
            this.minimizeButton.Click                      += (s, e) => this.WindowState = WindowState.Minimized;
            this.maximizeButton.Click                      += maximizeButton_Click;
            this.closeButton.Click                         += (s, e) => Exit();

            this.videoBorder.DragOver                      += videoBorder_DragOver;
            this.videoBorder.Drop                          += videoBorder_Drop;

            this.mediaElement.MediaOpened                  += mediaElement_MediaOpened;
            this.mediaElement.MediaEnded                   += mediaElement_MediaEnded;

            this.timeLineSlider.PreviewMouseMove           += timeLineSlider_PreviewMouseMove;
            this.timeLineSlider.PreviewMouseLeftButtonUp   += timeLineSlider_PreviewMouseLeftButtonUp;
            this.timeLineSlider.ValueChanged               += timeLineSlider_ValueChanged;
            this.muteToggleButton.Click                    += (s, e) => this.mediaElement.IsMuted = !this.mediaElement.IsMuted;

            this.playToggleButton.Click                    += playToggleButton_Click;
            this.stopButton.Click                          += (s, e) => { CloseMedia(); this.timeLineSlider.Value = 0; };
            this.previousButton.Click                      += (s, e) => PlayMedia(this.currentPlayIndex - 1);
            this.nextButton.Click                          += (s, e) => PlayMedia(this.currentPlayIndex + 1);
            this.openButton.Click                          += (s, e) => ShowFileDialog();
            this.playListButton.Click                      += playListButton_Click;

            this.timer.Tick                                += (s, e) => { this.timeLineSlider.Value = (int)Math.Round(this.mediaElement.Position.TotalSeconds); };

            this.mediaPlayCollection.CollectionChanged     += mediaPlayCollection_CollectionChanged;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 윈도우 키 DOWN 처리하기 - Window_KeyDown(sender, e)

        /// <summary>
        /// 윈도우 키 DOWN 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Window_KeyDown(object sender, KeyEventArgs e)
        {
            if(e.Key == Key.Escape)
            {
                if(this.isFullScreen)
                {
                    SetFullScreen();
                }
                else
                {
                    Exit();
                }
            }
        }

        #endregion
        #region 최대화 버튼 클릭시 처리하기 - maximizeButton_Click(sender, e)

        /// <summary>
        /// 최대화 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void maximizeButton_Click(object sender, RoutedEventArgs e)
        {
            if(WindowState == WindowState.Normal)
            {
                WindowState = WindowState.Maximized;
            }
            else
            {
                WindowState = WindowState.Normal;
            }
        }

        #endregion
        #region 비디오 테두리 드래그 OVER 처리하기 - videoBorder_DragOver(sender, e)

        /// <summary>
        /// 비디오 테두리 드래그 OVER 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void videoBorder_DragOver(object sender, DragEventArgs e)
        {
            if((e.AllowedEffects & DragDropEffects.Copy) == DragDropEffects.Copy)
            {
                if(e.Data.GetDataPresent(DataFormats.FileDrop))
                {
                    e.Effects = DragDropEffects.Copy;
                }
                else
                {
                    e.Effects = DragDropEffects.None;
                }

                e.Handled = true;
            }
        }

        #endregion
        #region 비디오 테두리 DROP 처리하기 - videoBorder_Drop(sender, e)

        /// <summary>
        /// 비디오 테두리 DROP 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void videoBorder_Drop(object sender, DragEventArgs e)
        {
            string[] filePathArray = e.Data.GetData(DataFormats.FileDrop) as string[];

            bool isClear = false;

            foreach(string filePath in filePathArray)
            {
                FileInfo fileInfo = new FileInfo(filePath);

                if(fileInfo.Extension.Equals(".wmv") || fileInfo.Extension.Equals(".avi") || fileInfo.Extension.Equals(".mp4"))
                {
                    if(!isClear)
                    {
                        isClear = true;

                        this.mediaPlayCollection.Clear();
                    }

                    this.mediaPlayCollection.Add
                    (
                        new MediaPlayListItem()
                        {
                            FileName = fileInfo.Name,
                            FilePath = filePath
                        }
                    );
                }
            }

            e.Handled = true;

            if(this.mediaPlayCollection.Count > 0)
            {
                PlayMedia(0);
            }
        }

        #endregion
        #region 미디어 엘리먼트 미디어 오픈시 처리하기 - mediaElement_MediaOpened(sender, e)

        /// <summary>
        /// 미디어 엘리먼트 미디어 오픈시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void mediaElement_MediaOpened(object sender, RoutedEventArgs e)
        {
            this.titleTextBlock.Text = this.mediaPlayCollection[this.currentPlayIndex].FileName;

            this.Width  += this.mediaElement.NaturalVideoWidth  -
                (this.videoBorder.ActualWidth  - (this.videoBorder.BorderThickness.Left + this.videoBorder.BorderThickness.Right ));
            this.Height += this.mediaElement.NaturalVideoHeight -
                (this.videoBorder.ActualHeight - (this.videoBorder.BorderThickness.Top  + this.videoBorder.BorderThickness.Bottom));

            if(this.mediaElement.NaturalDuration.HasTimeSpan)
            {
                this.timeLineSlider.Maximum = (int)this.mediaElement.NaturalDuration.TimeSpan.TotalSeconds;

                this.totalTimeTextBlock.Text = this.mediaElement.NaturalDuration.TimeSpan.ToString().Substring(0, 8);
            }

            this.timeLineSlider.IsEnabled = true;

            this.playToggleButton.IsChecked = false;

            this.isClosed = false;
        }

        #endregion
        #region 미디어 엘리먼트 미디어 종료시 처리하기 - mediaElement_MediaEnded(sender, e)

        /// <summary>
        /// 미디어 엘리먼트 미디어 종료시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void mediaElement_MediaEnded(object sender, RoutedEventArgs e)
        {
            if(this.currentPlayIndex + 1 < this.mediaPlayCollection.Count)
            {
                this.mediaElement.Source = new Uri(this.mediaPlayCollection[++this.currentPlayIndex].FilePath, UriKind.Absolute);

                SetMediaPlayListWindowIndex();
            }
            else
            {
                CloseMedia();
            }
        }

        #endregion
        #region 타임 라인 슬라이더 PREVIEW 마우스 이동시 처리하기 - timeLineSlider_PreviewMouseMove(sender, e)

        /// <summary>
        /// 타임 라인 슬라이더 PREVIEW 마우스 이동시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void timeLineSlider_PreviewMouseMove(object sender, MouseEventArgs e)
        {
            if(e.LeftButton == MouseButtonState.Pressed && !this.isTimelineMoving)
            {
                this.isTimelineMoving = true;

                if(!this.isClosed)
                {
                    this.timer.Stop();

                    this.mediaElement.Pause();
                }
            }
        }

        #endregion
        #region 타임 라인 슬라이더 PREVIEW 마우스 왼쪽 버튼 UP 처리하기 - timeLineSlider_PreviewMouseLeftButtonUp(sender, e)

        /// <summary>
        /// 타임 라인 슬라이더 PREVIEW 마우스 왼쪽 버튼 UP 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void timeLineSlider_PreviewMouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            this.isTimelineMoving = false;

            if(!this.isPause && !this.isClosed)
            {
                this.mediaElement.Position = TimeSpan.FromSeconds(this.timeLineSlider.Value);

                this.timer.Start();

                this.mediaElement.Play();
            }
        }

        #endregion
        #region 타임 라인 슬라이더 값 변경시 처리하기 - timeLineSlider_ValueChanged(sender, e)

        /// <summary>
        /// 타임 라인 슬라이더 값 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void timeLineSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            if(this.isTimelineMoving && !this.isClosed)
            {
                this.mediaElement.Position = TimeSpan.FromSeconds(e.NewValue);
            }

            this.currentTimeTextBlock.Text = TimeSpan.FromSeconds(e.NewValue).ToString().Substring(0, 8);
        }

        #endregion
        #region 재생 토글 버튼 클릭시 처리하기 - playToggleButton_Click(sender, e)

        /// <summary>
        /// 재생 토글 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void playToggleButton_Click(object sender, RoutedEventArgs e)
        {
            if(this.mediaElement.Source == null)
            {
                this.playToggleButton.IsChecked = true;

                ShowFileDialog();

                return;
            }

            if(this.isPause || this.isClosed)
            {
                PlayMedia();
            }
            else
            {
                PauseMedia();
            }
        }

        #endregion
        #region 재생 리스트 버튼 클릭시 처리하기 - playListButton_Click(sender, e)

        /// <summary>
        /// 재생 리스트 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void playListButton_Click(object sender, RoutedEventArgs e)
        {
            if(Application.Current.Windows.Count > 1)
            {
                Application.Current.Windows[1].Close();
            }
            else
            {
                MediaPlayListWindow window = new MediaPlayListWindow();

                window.ItemsSource = this.mediaPlayCollection;

                window.ListIndexChanged += (index) => PlayMedia(index);

                window.SelectedIndex = 0;

                if(WindowState == WindowState.Maximized)
                {
                    window.WindowStartupLocation = WindowStartupLocation.CenterScreen;
                }
                else
                {
                    window.Left = Left + ActualWidth;
                    window.Top  = Top;
                }

                window.Show();
            }
        }

        #endregion
        #region 미디어 재생 리스트 컬렉션 변경시 처리하기 - mediaPlayListCollection_CollectionChanged(sender, e)

        /// <summary>
        /// 미디어 재생 리스트 컬렉션 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void mediaPlayCollection_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if(e.Action == NotifyCollectionChangedAction.Remove)
            {
                if(this.mediaPlayCollection.Count > 0)
                {
                    PlayMedia(this.currentPlayIndex + 1);
                }
                else
                {
                    CloseMedia();

                    this.mediaElement.Source = null;

                    this.timeLineSlider.Value = 0;

                    this.totalTimeTextBlock.Text = "00:00:00";
                }
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Function

        #region 전체 화면 설정하기 - SetFullScreen()

        /// <summary>
        /// 전체 화면 설정하기
        /// </summary>
        private void SetFullScreen()
        {
            TimeSpan mediaElementPosition = this.mediaElement.Position;

            this.isFullScreen = !this.isFullScreen;

            if(this.isFullScreen)
            {
                this.videoBorder.Child = null;

                Content     = this.mediaElement;
                Background  = Brushes.Black;
                WindowState = WindowState.Maximized;
            }
            else
            {
                Content = this.rootGrid;

                this.videoBorder.Child = this.mediaElement;

                Background  = Brushes.White;
                WindowState = WindowState.Normal;
            }

            this.mediaElement.Position = mediaElementPosition;
        }

        #endregion
        #region 종료하기 - Exit()

        /// <summary>
        /// 종료하기
        /// </summary>
        private void Exit()
        {
            this.timer.Stop();

            this.mediaElement.Close();

            foreach(Window window in Application.Current.Windows)
            {
                window.Close();
            }
        }

        #endregion
        #region 파일 대화 상자 표시하기 - ShowFileDialog()

        /// <summary>
        /// 파일 대화 상자 표시하기
        /// </summary>
        private void ShowFileDialog()
        {
            OpenFileDialog openFileDialog = new OpenFileDialog()
            {
                Filter      = "미디어 파일 (*.wmv, *.avi, *.mp4)|*.wmv;*.avi;*.mp4",
                Multiselect = true
            };

            if((bool)openFileDialog.ShowDialog())
            {
                this.mediaPlayCollection.Clear();

                string[] filePathArray = openFileDialog.FileNames;
                string[] fileNameArray = openFileDialog.SafeFileNames;

                int i = 0;

                foreach(string filePath in filePathArray)
                {
                    this.mediaPlayCollection.Add
                    (
                        new MediaPlayListItem()
                        {
                            FilePath = filePath,
                            FileName = fileNameArray[i++]
                        }
                    );
                }

                PlayMedia(0);
            }
        }

        #endregion
        #region 미디어 재생 리스트 윈도우 인덱스 설정하기 - SetMediaPlayListWindowIndex()

        /// <summary>
        /// 미디어 재생 리스트 윈도우 인덱스 설정하기
        /// </summary>
        private void SetMediaPlayListWindowIndex()
        {
            if(Application.Current.Windows.Count > 1)
            {
                (Application.Current.Windows[1] as MediaPlayListWindow).SelectedIndex = this.currentPlayIndex;
            }
        }

        #endregion
        #region 미디어 재생하기 - PlayMedia()

        /// <summary>
        /// 미디어 재생하기
        /// </summary>
        private void PlayMedia()
        {
            if(this.isClosed)
            {
                this.timeLineSlider.Value = 0;
            }

            this.timer.Start();

            this.mediaElement.Play();

            this.isPause = false;
        }

        #endregion
        #region 미디어 재생하기 - PlayMedia(index)

        /// <summary>
        /// 미디어 재생하기
        /// </summary>
        /// <param name="index">인덱스</param>
        private void PlayMedia(int index)
        {
            if(this.mediaPlayCollection.Count.Equals(0))
            {
                return;
            }

            if(index >= this.mediaPlayCollection.Count)
            {
                index = 0;
            }
            else if(index < 0)
            {
                index = this.mediaPlayCollection.Count - 1;
            }

            this.currentPlayIndex = index;

            this.mediaElement.Source = new Uri(this.mediaPlayCollection[this.currentPlayIndex].FilePath, UriKind.Absolute);

            PlayMedia();

            SetMediaPlayListWindowIndex();
        }

        #endregion
        #region 미디어 정지하기 - PauseMedia()

        /// <summary>
        /// 미디어 정지하기
        /// </summary>
        private void PauseMedia()
        {
            this.timer.Stop();

            this.mediaElement.Pause();

            this.isPause = true;
        }

        #endregion
        #region 미디어 닫기 - CloseMedia()

        /// <summary>
        /// 미디어 닫기
        /// </summary>
        private void CloseMedia()
        {
            this.timer.Stop();

            this.mediaElement.Close();

            this.playToggleButton.IsChecked = true;

            this.isClosed = true;
        }

        #endregion
    }
}